import 'package:code_id_flutter/code_id_flutter.dart';
import 'package:code_id_flutter/code_interfaces/storage/token_storage.dart';
import 'package:code_id_flutter/code_services/services/network_service.dart';
import 'package:dio/dio.dart';

typedef ShouldRefresh = bool Function(Response? response);

typedef RefreshToken<T> = Future<T> Function(
  T? token,
  INetworkService httpClient,
);

typedef ShouldRequest<T> = Future<T> Function(
  T? token,
  INetworkService httpClient,
);

class RefreshTokenInterceptor<T> extends Interceptor with TokenMixin<T> {
  final INetworkService _httpClient;
  final TokenHeaderBuilder<T> _tokenHeader;
  final ShouldRefresh _shouldRefresh;
  final ShouldRequest<T>? _shouldRequest;
  final RefreshToken<T>? _refreshToken;

  RefreshTokenInterceptor({
    required TokenHeaderBuilder<T> tokenHeader,
    required TokenStorage<T> tokenStorage,
    RefreshToken<T>? refreshToken,

    /// Should Request for the request if you want to check if the T class have timeout Token.
    ShouldRequest<T>? shouldRequest,
    ShouldRefresh? shouldRefresh,
    INetworkService? httpClient,
  })  : _refreshToken = refreshToken,
        _tokenHeader = tokenHeader,
        _shouldRequest = shouldRequest,
        _shouldRefresh = shouldRefresh ?? _defaultShouldRefresh,
        _httpClient = httpClient ?? NetworkService() {
    this.tokenStorage = tokenStorage;
  }

  @override
  Future<dynamic> onRequest(
    RequestOptions options,
    RequestInterceptorHandler handler,
  ) async {
    T? currentToken;
    if (_shouldRequest != null) {
      currentToken = await _shouldRequest!(await token, _httpClient);
    } else {
      currentToken = await token;
    }
    final headers = currentToken != null
        ? _tokenHeader(currentToken)
        : const <String, String>{};
    options.headers.addAll(headers);
    handler.next(options);
  }

  @override
  Future<dynamic> onResponse(
    Response response,
    ResponseInterceptorHandler handler,
  ) async {
    if (await token == null || !_shouldRefresh(response)) {
      return handler.next(response);
    }
    try {
      final refreshResponse = await _tryRefresh(response);
      handler.resolve(refreshResponse);
    } on DioError catch (error) {
      handler.reject(error);
    }
  }

  @override
  Future<dynamic> onError(
    DioError err,
    ErrorInterceptorHandler handler,
  ) async {
    final response = err.response;
    if (response == null ||
        await token == null ||
        err.error is RevokeTokenException ||
        !_shouldRefresh(response)) {
      return handler.next(err);
    }
    try {
      final refreshResponse = await _tryRefresh(response);
      handler.resolve(refreshResponse);
    } on DioError catch (error) {
      handler.next(error);
    }
  }

  Future<Response<dynamic>> _tryRefresh(Response response) async {
    late final T refreshedToken;
    if (_refreshToken != null) {
      try {
        refreshedToken = await _refreshToken!(await token, _httpClient);
      } on RevokeTokenException catch (error) {
        await clearToken();
        throw DioError(
          requestOptions: response.requestOptions,
          error: error,
          response: response,
        );
      }

      await setToken(refreshedToken);
      final Dio _newDio = Dio();
      _newDio.options.baseUrl = response.requestOptions.baseUrl;
      return _newDio.request<dynamic>(
        response.requestOptions.path,
        cancelToken: response.requestOptions.cancelToken,
        data: response.requestOptions.data,
        onReceiveProgress: response.requestOptions.onReceiveProgress,
        onSendProgress: response.requestOptions.onSendProgress,
        queryParameters: response.requestOptions.queryParameters,
        options: Options(
          method: response.requestOptions.method,
          sendTimeout: response.requestOptions.sendTimeout,
          receiveTimeout: response.requestOptions.receiveTimeout,
          extra: response.requestOptions.extra,
          headers: response.requestOptions.headers
            ..addAll(_tokenHeader(refreshedToken)),
          responseType: response.requestOptions.responseType,
          contentType: response.requestOptions.contentType,
          validateStatus: response.requestOptions.validateStatus,
          receiveDataWhenStatusError:
              response.requestOptions.receiveDataWhenStatusError,
          followRedirects: response.requestOptions.followRedirects,
          maxRedirects: response.requestOptions.maxRedirects,
          requestEncoder: response.requestOptions.requestEncoder,
          responseDecoder: response.requestOptions.responseDecoder,
          listFormat: response.requestOptions.listFormat,
        ),
      );
    } else {
      throw NetworkError;
    }
  }

  static bool _defaultShouldRefresh(Response? response) {
    return response?.statusCode == 401;
  }
}
